# :coding: utf-8

import copy
import os
import types
from collections import OrderedDict, Counter

import pytest

import wiz.definition
import wiz.exception
import wiz.filesystem
import wiz.system
from wiz.utility import Requirement, Version


@pytest.fixture()
def definitions():
    """Return list of mocked definitions."""
    return [
        wiz.definition.Definition({
            "identifier": "foo",
            "namespace": "test",
            "version": "0.1.0",
        }),
        wiz.definition.Definition({
            "identifier": "foo",
            "namespace": "test",
            "version": "1.1.0",
            "command": {
                "foo": "Foo1.1",
            }
        }),
        wiz.definition.Definition({
            "identifier": "bar",
            "version": "1.0.0",
            "command": {
                "bar": "Bar1.0",
            }
        }),
        wiz.definition.Definition({
            "identifier": "bar",
            "version": "0.9.2",
            "command": {
                "bar": "Bar0.9",
            }
        }),
        wiz.definition.Definition({
            "identifier": "baz",
            "version": "0.1.1",
            "command": {
                "baz": "Baz0.1",
            }
        }),
        wiz.definition.Definition({
            "identifier": "bim",
            "version": "0.2.1",
            "command": {
                "bim": "Bim0.2",
            }
        }),
        wiz.definition.Definition({
            "identifier": "bim",
            "version": "0.2.1",
            "command": {
                "bim-test": "Bim0.2 --test",
            }
        }),
        wiz.definition.Definition({
            "identifier": "bim",
            "version": "0.1.0",
            "command": {
                "bim": "Bim0.1",
            }
        }),
        wiz.definition.Definition({
            "identifier": "foo",
            "namespace": "other"
        }),
        wiz.definition.Definition({
            "identifier": "foo",
        })
    ]


@pytest.fixture()
def registries(temporary_directory):
    """Return mocked registry paths."""
    mapping = {
        "registry1": {
            "__files__": ["defA.json"],
            "level1": {
                "__files__": ["defB.jee-son"],
                "level2": {
                    "__files__": ["defC.json", "defD"],
                    "level3": {
                        "__files__": ["defE.json", "defF.json"],
                    }
                }
            }
        },
        "registry2": {
            "__files__": ["defG.yml", "defH.json", "defI.json"],
        }
    }

    def _create_structure(root, _mapping):
        """Create the mocked registry structure *_mapping* in *root*."""
        for key, value in _mapping.items():
            if key == "__files__":
                for _file in value:
                    path = os.path.join(root, _file)
                    with open(path, "w") as stream:
                        stream.write("")

            else:
                path = os.path.join(root, key)
                os.makedirs(path)
                _create_structure(path, value)

    _create_structure(temporary_directory, mapping)

    return [
        os.path.join(temporary_directory, "registry1"),
        os.path.join(temporary_directory, "registry2"),
        " "
    ]


@pytest.fixture()
def mocked_system_validate(mocker):
    """Return mocked system.validate function."""
    return mocker.patch.object(wiz.system, "validate")


@pytest.fixture()
def mocked_filesystem_export(mocker):
    """Return mocked filesystem.export function."""
    return mocker.patch.object(wiz.filesystem, "export")


@pytest.fixture()
def mocked_system_validate(mocker):
    """Return mocked validate function."""
    return mocker.patch.object(wiz.system, "validate")


@pytest.fixture()
def mocked_discover(mocker):
    """Return mocked discovery function."""
    return mocker.patch.object(wiz.definition, "discover")


@pytest.fixture()
def mocked_load(mocker):
    """Return mocked load function."""
    return mocker.patch.object(wiz.definition, "load")


@pytest.fixture()
def mocked_fetch(mocker):
    """Return mocked fetch function."""
    return mocker.patch.object(wiz.definition, "fetch")


@pytest.fixture()
def mocked_query(mocker):
    """Return mocked query function."""
    return mocker.patch.object(wiz.definition, "query")


@pytest.fixture()
def mocked_definition(mocker):
    """Return mocked Definition class."""
    return mocker.patch.object(
        wiz.definition, "Definition", return_value="DEFINITION"
    )


@pytest.fixture()
def mocked_registry_install(mocker):
    """Return mocked load function."""
    return mocker.patch.object(wiz.registry, "install")


@pytest.mark.parametrize("options", [
    {},
    {"max_depth": 4},
    {"system_mapping": "__SYSTEM__"}
], ids=[
    "without-option",
    "with-max-depth",
    "with-system"
])
def test_fetch(mocked_discover, definitions, options):
    """Fetch all definition within *paths*."""
    mocked_discover.return_value = definitions
    result = wiz.definition.fetch(
        ["/path/to/registry-1", "/path/to/registry-2"], **options
    )

    mocked_discover.assert_called_once_with(
        ["/path/to/registry-1", "/path/to/registry-2"],
        max_depth=options.get("max_depth"),
        system_mapping=options.get("system_mapping")
    )

    assert result == {
        "package": {
            "__namespace__": {
                "foo": {"test", "other"}
            },
            "test::foo": {
                "0.1.0": definitions[0],
                "1.1.0": definitions[1]
            },
            "bar": {
                "1.0.0": definitions[2],
                "0.9.2": definitions[3]
            },
            "baz": {
                "0.1.1": definitions[4]
            },
            "bim": {
                # The 5th definition in the incoming list is overridden by the
                # 6th one which has the same identifier and version.
                "0.2.1": definitions[6],
                "0.1.0": definitions[7]
            },
            "other::foo": {
                "-": definitions[8]
            },
            "foo": {
                "-": definitions[9]
            }
        },
        "command": {
            "foo": "test::foo",
            "bar": "bar",
            "baz": "baz",
            "bim-test": "bim",
            "bim": "bim"
        },
        "implicit-packages": []
    }


@pytest.mark.parametrize("options", [
    {},
    {"max_depth": 4},
    {"system_mapping": "__SYSTEM__"}
], ids=[
    "without-option",
    "with-max-depth",
    "with-system"
])
def test_fetch_with_implicit_packages(mocked_discover, definitions, options):
    """Fetch all definition within *paths*."""
    definitions[0] = definitions[0].set("auto-use", True)
    definitions[1] = definitions[1].set("auto-use", True)
    definitions[3] = definitions[3].set("auto-use", True)
    definitions[8] = definitions[8].set("auto-use", True)

    mocked_discover.return_value = definitions
    result = wiz.definition.fetch(
        ["/path/to/registry-1", "/path/to/registry-2"], **options
    )

    mocked_discover.assert_called_once_with(
        ["/path/to/registry-1", "/path/to/registry-2"],
        max_depth=options.get("max_depth"),
        system_mapping=options.get("system_mapping"),
    )

    assert result == {
        "package": {
            "__namespace__": {
                "foo": {"test", "other"}
            },
            "test::foo": {
                "0.1.0": definitions[0],
                "1.1.0": definitions[1]
            },
            "bar": {
                "1.0.0": definitions[2],
                "0.9.2": definitions[3]
            },
            "baz": {
                "0.1.1": definitions[4]
            },
            "bim": {
                # The 5th definition in the incoming list is overridden by the
                # 6th one which has the same identifier and version.
                "0.2.1": definitions[6],
                "0.1.0": definitions[7]
            },
            "other::foo": {
                "-": definitions[8]
            },
            "foo": {
                "-": definitions[9]
            }
        },
        "command": {
            "foo": "test::foo",
            "bar": "bar",
            "baz": "baz",
            "bim-test": "bim",
            "bim": "bim"
        },
        "implicit-packages": [
            "other::foo",
            "bar==0.9.2",
            "test::foo==1.1.0"
        ]
    }


def test_query_definition():
    """Return best matching definition from requirement."""
    package_mapping = {
        "bar": {
            "0.3.0": wiz.definition.Definition({
                "identifier": "bar",
                "version": "0.3.0",
            }),
            "0.1.5": wiz.definition.Definition({
                "identifier": "bar",
                "version": "0.1.5",
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "bar",
                "version": "0.1.0",
            }),
        },
        "foo": {
            "1.1.0": wiz.definition.Definition({
                "identifier": "foo",
                "version": "1.1.0",
            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "foo",
                "version": "0.1.0",
            }),
        }
    }

    requirement = Requirement("bar")
    assert (
        wiz.definition.query(requirement, package_mapping) ==
        package_mapping["bar"]["0.3.0"]
    )

    requirement = Requirement("foo<0.2")
    assert (
        wiz.definition.query(requirement, package_mapping) ==
        package_mapping["foo"]["0.1.0"]
    )

    requirement = Requirement("bar==0.1.5")
    assert (
        wiz.definition.query(requirement, package_mapping) ==
        package_mapping["bar"]["0.1.5"]
    )


def test_query_definition_explicit_namespace():
    """Explicitly query package identifier with or without namespace."""
    package_mapping = {
        "__namespace__": {
            "foo": {"namespace"},
        },
        "foo": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "foo",
                "version": "0.1.0",
            }),
        },
        "namespace::foo": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "foo",
                "version": "0.1.0",
                "namespace": "namespace"
            }),
        },
    }

    requirement = Requirement("::foo")
    assert (
        wiz.definition.query(requirement, package_mapping) ==
        package_mapping["foo"]["0.1.0"]
    )

    requirement = Requirement("namespace::foo")
    assert (
        wiz.definition.query(requirement, package_mapping) ==
        package_mapping["namespace::foo"]["0.1.0"]
    )


def test_query_definition_guess_default_one_namespace():
    """Guess namespace when only one namespace is available for identifier."""
    package_mapping = {
        "__namespace__": {
            "foo": {"namespace"},
        },
        "namespace::foo": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "foo",
                "version": "0.1.0",
                "namespace": "namespace"
            }),
        },
    }

    requirement = Requirement("foo")
    assert (
        wiz.definition.query(requirement, package_mapping) ==
        package_mapping["namespace::foo"]["0.1.0"]
    )


def test_query_definition_guess_default_namespace_identifier():
    """Guess namespace when only one namespace is equal to identifier."""
    package_mapping = {
        "__namespace__": {
            "foo": {"namespace", "foo"},
        },
        "foo::foo": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "foo",
                "version": "0.1.0",
                "namespace": "foo"
            }),
        },
        "namespace::foo": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "foo",
                "version": "0.2.0",
                "namespace": "namespace"
            }),
        },
    }

    requirement = Requirement("foo")
    assert (
        wiz.definition.query(requirement, package_mapping) ==
        package_mapping["foo::foo"]["0.1.0"]
    )


def test_query_definition_guess_default_namespace_counter():
    """Guess namespace from namespace counter."""
    package_mapping = {
        "__namespace__": {
            "foo": {"namespace1", "namespace2"},
        },
        "namespace1::foo": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "foo",
                "version": "0.1.0",
                "namespace": "namespace1"
            }),
        },
        "namespace2::foo": {
            "0.2.0": wiz.definition.Definition({
                "identifier": "foo",
                "version": "0.2.0",
                "namespace": "namespace2"
            }),
        },
    }

    requirement = Requirement("foo")

    # No counter
    with pytest.raises(wiz.exception.RequestNotFound) as error:
        wiz.definition.query(requirement, package_mapping)

    assert (
        "Cannot guess default namespace for 'foo' "
        "[available: namespace1, namespace2]"
    ) in str(error)

    # Counter with "namespace1"
    assert (
        wiz.definition.query(
            requirement, package_mapping,
            namespace_counter=Counter(["namespace1", "namespace1"])
        ) ==
        package_mapping["namespace1::foo"]["0.1.0"]
    )

    assert (
        wiz.definition.query(
            requirement, package_mapping,
            namespace_counter=Counter(["namespace1"])
        ) ==
        package_mapping["namespace1::foo"]["0.1.0"]
    )

    # Counter with "namespace1" and "namespace2"
    with pytest.raises(wiz.exception.RequestNotFound) as error:
        wiz.definition.query(
            requirement, package_mapping,
            namespace_counter=Counter([
                "namespace1", "namespace2", "namespace1", "namespace2"
            ])
        )

    assert (
        "Cannot guess default namespace for 'foo' "
        "[available: namespace1, namespace2]"
    ) in str(error)


def test_query_definition_with_variant_identifier():
    """Query definition version from variant identifier."""
    package_mapping = {
        "foo": {
            "0.1.1": wiz.definition.Definition({
                "identifier": "foo",
                "version": "0.1.1",
                "variants": [
                    {
                        "identifier": "V2"
                    }
                ]

            }),
            "0.1.0": wiz.definition.Definition({
                "identifier": "foo",
                "version": "0.1.0",
                "variants": [
                    {
                        "identifier": "V1"
                    }
                ]
            }),
        },
    }

    requirement = Requirement("foo[V1]")
    assert (
        wiz.definition.query(requirement, package_mapping) ==
        package_mapping["foo"]["0.1.0"]
    )

    requirement = Requirement("foo[V2]")
    assert (
        wiz.definition.query(requirement, package_mapping) ==
        package_mapping["foo"]["0.1.1"]
    )


def test_query_definition_name_error():
    """Fails to query the definition name."""
    package_mapping = {}

    requirement = Requirement("incorrect")

    with pytest.raises(wiz.exception.RequestNotFound):
        wiz.definition.query(requirement, package_mapping)


def test_query_definition_version_error():
    """Fails to query the definition version."""
    package_mapping = {
        "foo": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "foo",
                "version": "0.1.0",
            }),
        },
    }

    requirement = Requirement("foo>10")

    with pytest.raises(wiz.exception.RequestNotFound):
        wiz.definition.query(requirement, package_mapping)


def test_query_definition_mixed_version_error():
    """Fails to query definition from non-versioned and versioned definitions.
    """
    package_mapping = {
        "foo": {
            "0.1.0": wiz.definition.Definition({
                "identifier": "foo",
                "version": "0.1.0",
            }),
            "-": wiz.definition.Definition({
                "identifier": "foo",
            }),
        },
    }

    requirement = Requirement("foo")

    with pytest.raises(wiz.exception.RequestNotFound) as error:
        wiz.definition.query(requirement, package_mapping)

    assert (
        "RequestNotFound: Impossible to retrieve the best matching definition "
        "for 'foo' as non-versioned and versioned definitions have been "
        "fetched."
    ) in str(error)


def test_export_data(mocked_filesystem_export):
    """Export definition data as a JSON file."""
    data = {
        "identifier": "foo",
        "description": "Test definition",
        "command": {
            "app": "App0.1",
            "appX": "AppX0.1"
        },
        "environ": {
            "KEY": "VALUE"
        }
    }

    wiz.definition.export("/path/to/output", data)

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/output/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"description\": \"Test definition\",\n"
            "    \"command\": {\n"
            "        \"app\": \"App0.1\",\n"
            "        \"appX\": \"AppX0.1\"\n"
            "    },\n"
            "    \"environ\": {\n"
            "        \"KEY\": \"VALUE\"\n"
            "    }\n"
            "}"
        ), overwrite=False
    )


def test_export_data_with_version(mocked_filesystem_export):
    """Export a definition data as a JSON file with version."""
    data = {
        "identifier": "foo",
        "version": "0.1.0",
        "description": "Test definition",
        "command": {
            "app": "App0.1",
            "appX": "AppX0.1"
        },
        "environ": {
            "KEY": "VALUE"
        }
    }

    wiz.definition.export("/path/to/output", data)

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/output/foo-0.1.0.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"version\": \"0.1.0\",\n"
            "    \"description\": \"Test definition\",\n"
            "    \"command\": {\n"
            "        \"app\": \"App0.1\",\n"
            "        \"appX\": \"AppX0.1\"\n"
            "    },\n"
            "    \"environ\": {\n"
            "        \"KEY\": \"VALUE\"\n"
            "    }\n"
            "}"
        ), overwrite=False
    )


def test_export(mocked_filesystem_export):
    """Export definition as a JSON file."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "description": "Test definition",
        "command": {
            "app": "App0.1",
            "appX": "AppX0.1"
        },
        "environ": {
            "KEY": "VALUE"
        }
    })

    wiz.definition.export("/path/to/output", definition)

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/output/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"description\": \"Test definition\",\n"
            "    \"command\": {\n"
            "        \"app\": \"App0.1\",\n"
            "        \"appX\": \"AppX0.1\"\n"
            "    },\n"
            "    \"environ\": {\n"
            "        \"KEY\": \"VALUE\"\n"
            "    }\n"
            "}"
        ), overwrite=False
    )


def test_export_with_version(mocked_filesystem_export):
    """Export a definition as a JSON file with version."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.1.0",
        "description": "Test definition",
        "command": {
            "app": "App0.1",
            "appX": "AppX0.1"
        },
        "environ": {
            "KEY": "VALUE"
        }
    })

    wiz.definition.export("/path/to/output", definition)

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/output/foo-0.1.0.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"version\": \"0.1.0\",\n"
            "    \"description\": \"Test definition\",\n"
            "    \"command\": {\n"
            "        \"app\": \"App0.1\",\n"
            "        \"appX\": \"AppX0.1\"\n"
            "    },\n"
            "    \"environ\": {\n"
            "        \"KEY\": \"VALUE\"\n"
            "    }\n"
            "}"
        ), overwrite=False
    )


def test_discover(mocked_load, mocked_system_validate, registries, definitions):
    """Discover and yield *definitions*."""
    mocked_load.side_effect = definitions

    result = wiz.definition.discover(registries)
    assert isinstance(result, types.GeneratorType)
    assert mocked_load.call_count == 0

    discovered = list(result)
    assert len(discovered) == 6
    assert mocked_load.call_count == 6

    r1 = registries[0]
    r2 = registries[1]

    path = os.path.join(r1, "defA.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r1, "level1", "level2", "defC.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r1, "level1", "level2", "level3", "defF.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r1, "level1", "level2", "level3", "defE.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r2, "defH.json")
    mocked_load.assert_any_call(
        path, registry_path=r2
    )

    path = os.path.join(r2, "defI.json")
    mocked_load.assert_any_call(
        path, registry_path=r2
    )

    assert discovered == definitions[:6]
    mocked_system_validate.assert_not_called()


def test_discover_with_max_depth(
    mocked_load, mocked_system_validate, registries, definitions
):
    """Discover and yield definitions with maximum depth."""
    mocked_load.side_effect = definitions

    result = wiz.definition.discover(registries, max_depth=2)
    assert isinstance(result, types.GeneratorType)
    assert mocked_load.call_count == 0

    discovered = list(result)
    assert len(discovered) == 4
    assert mocked_load.call_count == 4

    r1 = registries[0]
    r2 = registries[1]

    path = os.path.join(r1, "defA.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r1, "level1", "level2", "defC.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r2, "defH.json")
    mocked_load.assert_any_call(
        path, registry_path=r2
    )

    path = os.path.join(r2, "defI.json")
    mocked_load.assert_any_call(
        path, registry_path=r2
    )

    assert discovered == definitions[:4]
    mocked_system_validate.assert_not_called()


def test_discover_with_system_valid(
    mocked_load, mocked_system_validate, registries, definitions
):
    """Discover and yield definitions with valid system."""
    mocked_load.side_effect = definitions
    mocked_system_validate.return_value = True

    result = wiz.definition.discover(registries, system_mapping="__SYSTEM__")
    assert isinstance(result, types.GeneratorType)
    assert mocked_load.call_count == 0

    discovered = list(result)
    assert len(discovered) == 6
    assert mocked_load.call_count == 6

    r1 = registries[0]
    r2 = registries[1]

    path = os.path.join(r1, "defA.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r1, "level1", "level2", "defC.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r1, "level1", "level2", "level3", "defF.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r1, "level1", "level2", "level3", "defE.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r2, "defH.json")
    mocked_load.assert_any_call(
        path, registry_path=r2
    )

    path = os.path.join(r2, "defI.json")
    mocked_load.assert_any_call(
        path, registry_path=r2
    )

    assert discovered == definitions[:6]


def test_discover_with_system_invalid(
    mocked_load, mocked_system_validate, registries, definitions
):
    """Discover and yield definitions with invalid system."""
    mocked_load.side_effect = definitions
    mocked_system_validate.return_value = False

    result = wiz.definition.discover(registries, system_mapping="__SYSTEM__")
    assert isinstance(result, types.GeneratorType)
    assert mocked_load.call_count == 0

    discovered = list(result)
    assert len(discovered) == 0
    assert mocked_load.call_count == 6

    r1 = registries[0]
    r2 = registries[1]

    path = os.path.join(r1, "defA.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r1, "level1", "level2", "defC.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r1, "level1", "level2", "level3", "defF.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r1, "level1", "level2", "level3", "defE.json")
    mocked_load.assert_any_call(
        path, registry_path=r1
    )

    path = os.path.join(r2, "defH.json")
    mocked_load.assert_any_call(
        path, registry_path=r2
    )

    path = os.path.join(r2, "defI.json")
    mocked_load.assert_any_call(
        path, registry_path=r2
    )

    assert discovered == []


def test_discover_without_disabled(mocked_load, registries, definitions):
    """Discover and yield definitions without disabled definition."""
    data = definitions[2].data()
    data["disabled"] = True
    definitions[2] = wiz.definition.Definition(data)

    data = definitions[4].data()
    data["disabled"] = True
    definitions[4] = wiz.definition.Definition(data)

    mocked_load.side_effect = definitions

    result = wiz.definition.discover(registries)
    assert isinstance(result, types.GeneratorType)
    assert mocked_load.call_count == 0

    discovered = list(result)
    assert len(discovered) == 4
    assert mocked_load.call_count == 6

    assert discovered == [
        definitions[0],
        definitions[1],
        definitions[3],
        definitions[5]
    ]


@pytest.mark.parametrize("exception", [
    IOError,
    ValueError,
    TypeError,
    wiz.exception.WizError,
    wiz.exception.DefinitionError
], ids=[
    "io-error",
    "value-error",
    "type-error",
    "generic-wiz-error",
    "incorrect-definition",
])
def test_discover_skip_error(mocked_load, registries, exception):
    """Discover and yield definitions without definition with error."""
    mocked_load.side_effect = exception

    result = wiz.definition.discover(registries)
    assert isinstance(result, types.GeneratorType)
    assert mocked_load.call_count == 0

    discovered = list(result)
    assert len(discovered) == 0
    assert mocked_load.call_count == 6


@pytest.mark.parametrize("exception", [
    RuntimeError,
    Exception
], ids=[
    "runtime-error",
    "generic-error"
])
def test_discover_error(mocked_load, registries, exception):
    """Fail to discover and yield definitions."""
    mocked_load.side_effect = exception

    with pytest.raises(exception):
        list(wiz.definition.discover(registries))


def test_load(mocked_definition, temporary_file):
    """Load a definition from a path."""
    with open(temporary_file, "w") as stream:
        stream.write("{\"identifier\": \"test_definition\"}")

    mocked_definition.return_value = "DEFINITION"
    result = wiz.definition.load(temporary_file)
    assert result == "DEFINITION"
    mocked_definition.assert_called_once_with(
        {"identifier": "test_definition"},
        path=temporary_file,
        registry_path=None,
        copy_data=False
    )


def test_load_with_mapping(mocked_definition, temporary_file):
    """Load a definition from a path and a mapping."""
    with open(temporary_file, "w") as stream:
        stream.write("{\"identifier\": \"test_definition\"}")

    mocked_definition.return_value = "DEFINITION"
    result = wiz.definition.load(temporary_file, mapping={"key": "value"})
    assert result == "DEFINITION"
    mocked_definition.assert_called_once_with(
        {"identifier": "test_definition", "key": "value"},
        path=temporary_file,
        registry_path=None,
        copy_data=False
    )


def test_minimal_definition():
    """Create a minimal definition."""
    data = {"identifier": "test"}

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "test"
    assert definition.description is None
    assert definition.namespace is None
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\"\n"
        "}"
    )


def test_minimal_definition_with_paths():
    """Create a minimal definition with paths."""
    data = {"identifier": "test"}

    definition = wiz.definition.Definition(
        data,
        path="/path/to/definition.json",
        registry_path="/path/to/registry",
    )
    assert definition.path == "/path/to/definition.json"
    assert definition.registry_path == "/path/to/registry"
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "test"
    assert definition.description is None
    assert definition.namespace is None
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\"\n"
        "}"
    )


def test_minimal_definition_with_namespace():
    """Create a minimal definition with namespace."""
    data = {
        "identifier": "test",
        "namespace": "foo",
    }

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "foo::test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "foo::test"
    assert definition.description is None
    assert definition.namespace == "foo"
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("namespace", "foo"),
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"namespace\": \"foo\"\n"
        "}"
    )


def test_definition_with_version():
    """Create a definition with version."""
    data = {
        "identifier": "test",
        "version": "0.1.0",
    }

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version == Version("0.1.0")
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test==0.1.0"
    assert definition.qualified_version_identifier == "test==0.1.0"
    assert definition.description is None
    assert definition.namespace is None
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("version", "0.1.0"),
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"version\": \"0.1.0\"\n"
        "}"
    )


def test_definition_with_version_and_namespace():
    """Create a definition with version and namespace."""
    data = {
        "identifier": "test",
        "namespace": "foo",
        "version": "0.1.0",
    }

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version == Version("0.1.0")
    assert definition.qualified_identifier == "foo::test"
    assert definition.version_identifier == "test==0.1.0"
    assert definition.qualified_version_identifier == "foo::test==0.1.0"
    assert definition.description is None
    assert definition.namespace == "foo"
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("version", "0.1.0"),
        ("namespace", "foo"),
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"version\": \"0.1.0\",\n"
        "    \"namespace\": \"foo\"\n"
        "}"
    )


def test_definition_with_description():
    """Create a definition with description."""
    data = {
        "identifier": "test",
        "description": "This is a definition"
    }

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "test"
    assert definition.description == "This is a definition"
    assert definition.namespace is None
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("description", "This is a definition")
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"description\": \"This is a definition\"\n"
        "}"
    )


def test_definition_with_install_keywords():
    """Create a definition with installation keywords."""
    data = {
        "identifier": "test",
        "install-root": "/path/to/root",
        "install-location": "${INSTALL_ROOT}/install"
    }
    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "test"
    assert definition.description is None
    assert definition.namespace is None
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root == "/path/to/root"
    assert definition.install_location == "${INSTALL_ROOT}/install"
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("install-root", "/path/to/root"),
        ("install-location", "${INSTALL_ROOT}/install")
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"install-root\": \"/path/to/root\",\n"
        "    \"install-location\": \"${INSTALL_ROOT}/install\"\n"
        "}"
    )


def test_definition_with_auto_use():
    """Create a definition with 'auto-use' value."""
    data = {
        "identifier": "test",
        "auto-use": True
    }

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "test"
    assert definition.description is None
    assert definition.namespace is None
    assert definition.auto_use is True
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("auto-use", True)
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"auto-use\": true\n"
        "}"
    )


def test_definition_with_disabled():
    """Create a definition with 'disabled' value."""
    data = {
        "identifier": "test",
        "disabled": True
    }

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "test"
    assert definition.description is None
    assert definition.namespace is None
    assert definition.auto_use is False
    assert definition.disabled is True
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("disabled", True)
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"disabled\": true\n"
        "}"
    )


def test_definition_with_system():
    """Create a definition with system constraint."""
    data = {
        "identifier": "test",
        "system": {
            "arch": "x86_64",
            "platform": "linux"
        }
    }

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "test"
    assert definition.description is None
    assert definition.namespace is None
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {
        "arch": "x86_64",
        "platform": "linux"
    }
    assert definition.requirements == []
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("system", OrderedDict([
            ("platform", "linux"),
            ("arch", "x86_64"),
        ]))
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"system\": {\n"
        "        \"platform\": \"linux\",\n"
        "        \"arch\": \"x86_64\"\n"
        "    }\n"
        "}"
    )


def test_definition_with_command():
    """Create a definition with command."""
    data = {
        "identifier": "test",
        "command": {
            "app1": "App1",
            "app3": "App3",
            "app2": "App2",
        }
    }

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "test"
    assert definition.description is None
    assert definition.namespace is None
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {
        "app1": "App1",
        "app3": "App3",
        "app2": "App2",
    }
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("command", OrderedDict([
            ("app1", "App1"),
            ("app2", "App2"),
            ("app3", "App3"),
        ]))
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"command\": {\n"
        "        \"app1\": \"App1\",\n"
        "        \"app2\": \"App2\",\n"
        "        \"app3\": \"App3\"\n"
        "    }\n"
        "}"
    )


def test_definition_with_environ():
    """Create a definition with environment mapping."""
    data = {
        "identifier": "test",
        "environ": {
            "KEY1": "VALUE1",
            "KEY3": "VALUE3",
            "KEY2": "VALUE2",
        }
    }

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "test"
    assert definition.description is None
    assert definition.namespace is None
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {
        "KEY1": "VALUE1",
        "KEY3": "VALUE3",
        "KEY2": "VALUE2",
    }
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("environ", OrderedDict([
            ("KEY1", "VALUE1"),
            ("KEY2", "VALUE2"),
            ("KEY3", "VALUE3"),
        ]))
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"environ\": {\n"
        "        \"KEY1\": \"VALUE1\",\n"
        "        \"KEY2\": \"VALUE2\",\n"
        "        \"KEY3\": \"VALUE3\"\n"
        "    }\n"
        "}"
    )


def test_definition_with_requirements():
    """Create a definition with requirements."""
    data = {
        "identifier": "test",
        "requirements": [
            "envA >= 1.0.0",
            "envB >= 3.4.2, < 4",
            "envC"
        ]
    }

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "test"
    assert definition.description is None
    assert definition.namespace is None
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == [
        Requirement("envA >= 1.0.0"),
        Requirement("envB >= 3.4.2, < 4"),
        Requirement("envC")
    ]
    assert definition.conditions == []
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("requirements", [
            "envA >= 1.0.0",
            "envB >= 3.4.2, < 4",
            "envC"
        ])
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"requirements\": [\n"
        "        \"envA >= 1.0.0\",\n"
        "        \"envB >= 3.4.2, < 4\",\n"
        "        \"envC\"\n"
        "    ]\n"
        "}"
    )


def test_definition_with_conditions():
    """Create a definition with conditions."""
    data = {
        "identifier": "test",
        "conditions": [
            "envA >= 1.0.0",
            "envB >= 3.4.2, < 4",
            "envC"
        ]
    }

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "test"
    assert definition.description is None
    assert definition.namespace is None
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == [
        Requirement("envA >= 1.0.0"),
        Requirement("envB >= 3.4.2, < 4"),
        Requirement("envC")
    ]
    assert definition.variants == []

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("conditions", [
            "envA >= 1.0.0",
            "envB >= 3.4.2, < 4",
            "envC"
        ])
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"conditions\": [\n"
        "        \"envA >= 1.0.0\",\n"
        "        \"envB >= 3.4.2, < 4\",\n"
        "        \"envC\"\n"
        "    ]\n"
        "}"
    )


def test_definition_with_variants():
    """Create a definition with variants."""
    data = {
        "identifier": "test",
        "variants": [
            {
                "identifier": "V2",
                "install-location": "/path/to/V2",
                "environ": {
                    "KEY21": "VALUE21",
                    "KEY23": "VALUE23",
                    "KEY22": "VALUE22",
                },
                "command": {
                    "appXV2": "AppXV2",
                    "appV2": "AppV2",
                },
                "requirements": [
                    "envA >= 2, < 3"
                ]
            },
            {
                "identifier": "V1",
                "install-location": "/path/to/V1",
                "environ": {
                    "KEY11": "VALUE11",
                    "KEY13": "VALUE13",
                    "KEY12": "VALUE12",
                },
                "command": {
                    "appXV1": "AppXV1",
                    "appV1": "AppV1",
                },
                "requirements": [
                    "envA >= 1, < 2"
                ]
            }
        ]
    }

    definition = wiz.definition.Definition(data)
    assert definition.path is None
    assert definition.registry_path is None
    assert definition.identifier == "test"
    assert definition.version is None
    assert definition.qualified_identifier == "test"
    assert definition.version_identifier == "test"
    assert definition.qualified_version_identifier == "test"
    assert definition.description is None
    assert definition.namespace is None
    assert definition.auto_use is False
    assert definition.disabled is False
    assert definition.install_root is None
    assert definition.install_location is None
    assert definition.environ == {}
    assert definition.command == {}
    assert definition.system == {}
    assert definition.requirements == []
    assert definition.conditions == []

    assert len(definition.variants) == 2
    assert definition.variants[0].definition_identifier == "test"
    assert definition.variants[0].identifier == "V2"
    assert definition.variants[0].install_location == "/path/to/V2"
    assert definition.variants[0].environ == {
        "KEY21": "VALUE21",
        "KEY23": "VALUE23",
        "KEY22": "VALUE22",
    }
    assert definition.variants[0].command == {
        "appXV2": "AppXV2",
        "appV2": "AppV2",
    }
    assert definition.variants[0].requirements == [Requirement("envA >=2, <3")]

    assert definition.variants[1].definition_identifier == "test"
    assert definition.variants[1].identifier == "V1"
    assert definition.variants[1].install_location == "/path/to/V1"
    assert definition.variants[1].environ == {
        "KEY11": "VALUE11",
        "KEY13": "VALUE13",
        "KEY12": "VALUE12",
    }
    assert definition.variants[1].command == {
        "appXV1": "AppXV1",
        "appV1": "AppV1",
    }
    assert definition.variants[1].requirements == [Requirement("envA >=1, <2")]

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("variants", [
            OrderedDict([
                ("identifier", "V2"),
                ("install-location", "/path/to/V2"),
                ("command", OrderedDict([
                    ("appV2", "AppV2"),
                    ("appXV2", "AppXV2"),
                 ])),
                ("environ", OrderedDict([
                    ("KEY21", "VALUE21"),
                    ("KEY22", "VALUE22"),
                    ("KEY23", "VALUE23"),
                 ])),
                ("requirements", ["envA >= 2, < 3"])
            ]),
            OrderedDict([
                ("identifier", "V1"),
                ("install-location", "/path/to/V1"),
                ("command", OrderedDict([
                    ("appV1", "AppV1"),
                    ("appXV1", "AppXV1"),
                 ])),
                ("environ", OrderedDict([
                    ("KEY11", "VALUE11"),
                    ("KEY12", "VALUE12"),
                    ("KEY13", "VALUE13"),
                ])),
                ("requirements", ["envA >= 1, < 2"])
            ])
        ])
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"variants\": [\n"
        "        {\n"
        "            \"identifier\": \"V2\",\n"
        "            \"install-location\": \"/path/to/V2\",\n"
        "            \"command\": {\n"
        "                \"appV2\": \"AppV2\",\n"
        "                \"appXV2\": \"AppXV2\"\n"
        "            },\n"
        "            \"environ\": {\n"
        "                \"KEY21\": \"VALUE21\",\n"
        "                \"KEY22\": \"VALUE22\",\n"
        "                \"KEY23\": \"VALUE23\"\n"
        "            },\n"
        "            \"requirements\": [\n"
        "                \"envA >= 2, < 3\"\n"
        "            ]\n"
        "        },\n"
        "        {\n"
        "            \"identifier\": \"V1\",\n"
        "            \"install-location\": \"/path/to/V1\",\n"
        "            \"command\": {\n"
        "                \"appV1\": \"AppV1\",\n"
        "                \"appXV1\": \"AppXV1\"\n"
        "            },\n"
        "            \"environ\": {\n"
        "                \"KEY11\": \"VALUE11\",\n"
        "                \"KEY12\": \"VALUE12\",\n"
        "                \"KEY13\": \"VALUE13\"\n"
        "            },\n"
        "            \"requirements\": [\n"
        "                \"envA >= 1, < 2\"\n"
        "            ]\n"
        "        }\n"
        "    ]\n"
        "}"
    )


def test_definition_complete():
    """Create a definition with as many keywords as possible."""
    data = {
        "identifier": "test",
        "version": "0.1.0",
        "namespace": "foo",
        "description": "This is a definition",
        "auto-use": True,
        "install-root": "/path/to/root",
        "system": {
            "platform": "linux",
            "os": "el >= 6, < 7",
            "arch": "x86_64"
        },
        "command": {
            "app": "AppX"
        },
        "environ": {
            "KEY1": "VALUE1"
        },
        "requirements": ["envB"],
        "conditions": ["envC"],
        "variants": [
            {
                "identifier": "V2",
                "install-location": "${INSTALL_ROOT}/V2/install",
                "environ": {"KEY2": "VALUE2"},
                "command": {"appV2": "AppV2"},
                "requirements": [
                    "envA >= 2, < 3"
                ]
            }
        ]
    }

    definition = wiz.definition.Definition(
        data,
        path="/path/to/definition.json",
        registry_path="/path/to/registry",
    )
    assert definition.path == "/path/to/definition.json"
    assert definition.registry_path == "/path/to/registry"
    assert definition.identifier == "test"
    assert definition.version == Version("0.1.0")
    assert definition.qualified_identifier == "foo::test"
    assert definition.version_identifier == "test==0.1.0"
    assert definition.qualified_version_identifier == "foo::test==0.1.0"
    assert definition.description == "This is a definition"
    assert definition.namespace == "foo"
    assert definition.auto_use is True
    assert definition.disabled is False
    assert definition.install_root == "/path/to/root"
    assert definition.install_location is None
    assert definition.environ == {"KEY1": "VALUE1"}
    assert definition.command == {"app": "AppX"}
    assert definition.system == {
        "platform": "linux",
        "os": "el >= 6, < 7",
        "arch": "x86_64"
    }
    assert definition.requirements == [Requirement("envB")]
    assert definition.conditions == [Requirement("envC")]

    assert len(definition.variants) == 1
    assert definition.variants[0].definition_identifier == "test"
    assert definition.variants[0].identifier == "V2"
    assert definition.variants[0].install_location == (
        "${INSTALL_ROOT}/V2/install"
    )
    assert definition.variants[0].environ == {"KEY2": "VALUE2"}
    assert definition.variants[0].command == {"appV2": "AppV2"}
    assert definition.variants[0].requirements == [Requirement("envA >=2, <3")]

    assert definition.data() == data
    assert definition.ordered_data() == OrderedDict([
        ("identifier", "test"),
        ("version", "0.1.0"),
        ("namespace", "foo"),
        ("description", "This is a definition"),
        ("install-root", "/path/to/root"),
        ("auto-use", True),
        ("system", OrderedDict([
            ("platform", "linux"),
            ("os", "el >= 6, < 7"),
            ("arch", "x86_64"),
        ])),
        ("command", {"app": "AppX"}),
        ("environ", {"KEY1": "VALUE1"}),
        ("requirements", ["envB"]),
        ("conditions", ["envC"]),
        ("variants", [
            OrderedDict([
                ("identifier", "V2"),
                ("install-location", "${INSTALL_ROOT}/V2/install"),
                ("command", {"appV2": "AppV2"}),
                ("environ", {"KEY2": "VALUE2"}),
                ("requirements", ["envA >= 2, < 3"])
            ]),
        ])
    ])
    assert definition.encode() == (
        "{\n"
        "    \"identifier\": \"test\",\n"
        "    \"version\": \"0.1.0\",\n"
        "    \"namespace\": \"foo\",\n"
        "    \"description\": \"This is a definition\",\n"
        "    \"install-root\": \"/path/to/root\",\n"
        "    \"auto-use\": true,\n"
        "    \"system\": {\n"
        "        \"platform\": \"linux\",\n"
        "        \"os\": \"el >= 6, < 7\",\n"
        "        \"arch\": \"x86_64\"\n"
        "    },\n"
        "    \"command\": {\n"
        "        \"app\": \"AppX\"\n"
        "    },\n"
        "    \"environ\": {\n"
        "        \"KEY1\": \"VALUE1\"\n"
        "    },\n"
        "    \"requirements\": [\n"
        "        \"envB\"\n"
        "    ],\n"
        "    \"conditions\": [\n"
        "        \"envC\"\n"
        "    ],\n"
        "    \"variants\": [\n"
        "        {\n"
        "            \"identifier\": \"V2\",\n"
        "            \"install-location\": \"${INSTALL_ROOT}/V2/install\",\n"
        "            \"command\": {\n"
        "                \"appV2\": \"AppV2\"\n"
        "            },\n"
        "            \"environ\": {\n"
        "                \"KEY2\": \"VALUE2\"\n"
        "            },\n"
        "            \"requirements\": [\n"
        "                \"envA >= 2, < 3\"\n"
        "            ]\n"
        "        }\n"
        "    ]\n"
        "}"
    )


def test_definition_without_data_copy():
    """Create definition with copy input data."""
    data = {"identifier": "test"}
    definition = wiz.definition.Definition(data)

    data["key"] = "other"
    assert data != definition.data()

    data = {"identifier": "test"}
    definition = wiz.definition.Definition(data, copy_data=False)

    data["key"] = "other"
    assert data == definition.data()


def test_definition_data():
    """Fetch data from definition."""
    definition = wiz.definition.Definition({"identifier": "test"})
    data = definition.data()
    data["key"] = "other"

    assert data != definition.data()

    definition = wiz.definition.Definition({"identifier": "test"})
    data = definition.data(copy_data=False)
    data["key"] = "other"

    assert data == definition.data()


def test_definition_with_error():
    """Fail to create a definition with error."""
    data = {}

    with pytest.raises(wiz.exception.DefinitionError) as error:
        wiz.definition.Definition(data)

    assert "'identifier' is required." in str(error)


def test_definition_with_version_error():
    """Fail to create a definition with incorrect version."""
    data = {
        "identifier": "test",
        "version": "!!!"
    }

    with pytest.raises(wiz.exception.DefinitionError) as error:
        wiz.definition.Definition(data)

    assert "Invalid version: '!!!'" in str(error)


def test_definition_with_requirement_error():
    """Fail to create a definition with incorrect requirement."""
    data = {
        "identifier": "test",
        "requirements": [
            "envA -!!!",
        ]
    }

    definition = wiz.definition.Definition(data)

    with pytest.raises(wiz.exception.RequirementError) as error:
        print(definition.requirements)

    assert "The requirement 'envA -!!!' is incorrect" in str(error)


def test_definition_with_condition_error():
    """Fail to create a definition with incorrect condition."""
    data = {
        "identifier": "test",
        "conditions": [
            "envA -!!!",
        ]
    }

    definition = wiz.definition.Definition(data)

    with pytest.raises(wiz.exception.RequirementError) as error:
        print(definition.conditions)

    assert "The requirement 'envA -!!!' is incorrect" in str(error)


def test_definition_with_variant_requirement_error():
    """Fail to create a definition with incorrect variant requirement."""
    data = {
        "identifier": "test",
        "variants": [
            {
                "identifier": "1.0",
                "requirements": [
                    "envA -!!!"
                ]
            }
        ]
    }

    definition = wiz.definition.Definition(data)

    with pytest.raises(wiz.exception.RequirementError) as error:
        print(definition.variants[0].requirements)

    assert "The requirement 'envA -!!!' is incorrect" in str(error)


def test_definition_set():
    """Create new definition from existing definition with new element."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
    })

    assert definition1.data() == {"identifier": "foo"}

    # Change identifier
    definition2 = definition1.set("identifier", "bar")
    assert definition2.data() == {"identifier": "bar"}
    assert definition1.data() == {"identifier": "foo"}

    # Add version
    definition3 = definition2.set("version", "0.1.0")
    assert definition3.data() == {
        "identifier": "bar",
        "version": "0.1.0"
    }
    assert definition2.data() == {"identifier": "bar"}
    assert definition1.data() == {"identifier": "foo"}


def test_definition_update():
    """Create new definition from existing definition with updated element."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
    })

    assert definition1.data() == {"identifier": "foo"}

    definition2 = definition1.update(
        "environ", {"key1": "value1", "key2": "value2"}
    )
    assert definition2.data() == {
        "identifier": "foo",
        "environ": {
            "key1": "value1",
            "key2": "value2"
        }
    }
    assert definition1.data() == {"identifier": "foo"}

    definition3 = definition2.update(
        "environ", {"key1": "VALUE1", "key3": "value3"}
    )
    assert definition3.data() == {
        "identifier": "foo",
        "environ": {
            "key1": "VALUE1",
            "key2": "value2",
            "key3": "value3"
        }
    }
    assert definition2.data() == {
        "identifier": "foo",
        "environ": {
            "key1": "value1",
            "key2": "value2"
        }
    }
    assert definition1.data() == {"identifier": "foo"}


def test_definition_update_error():
    """Fail to create new definition with non-dictionary element updated."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
    })

    assert definition1.data() == {"identifier": "foo"}

    with pytest.raises(ValueError):
        definition1.update("identifier", {"key1": "value1"})


def test_definition_extend():
    """Create new definition from existing definition with extended element."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
    })

    assert definition1.data() == {"identifier": "foo"}

    definition2 = definition1.extend(
        "requirements", ["bar", "bim >= 1"]
    )
    assert definition2.data() == {
        "identifier": "foo",
        "requirements": ["bar", "bim >= 1"]
    }
    assert definition1.data() == {"identifier": "foo"}

    definition3 = definition2.extend(
        "requirements", ["test"]
    )
    assert definition3.data() == {
        "identifier": "foo",
        "requirements": ["bar", "bim >= 1", "test"]
    }
    assert definition2.data() == {
        "identifier": "foo",
        "requirements": ["bar", "bim >= 1"]
    }
    assert definition1.data() == {"identifier": "foo"}


def test_definition_extend_error():
    """Fail to create new definition with non-list element extended."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
    })

    assert definition1.data() == {"identifier": "foo"}

    with pytest.raises(ValueError):
        definition1.extend("identifier", ["test"])


def test_definition_insert():
    """Create new definition from existing definition with extended element."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
    })

    assert definition1.data() == {"identifier": "foo"}

    definition2 = definition1.set(
        "requirements", ["bar", "bim >= 1"]
    )
    assert definition2.data() == {
        "identifier": "foo",
        "requirements": ["bar", "bim >= 1"]
    }
    assert definition1.data() == {"identifier": "foo"}

    definition3 = definition2.insert(
        "requirements", "test", 0
    )
    assert definition3.data() == {
        "identifier": "foo",
        "requirements": ["test", "bar", "bim >= 1"]
    }
    assert definition2.data() == {
        "identifier": "foo",
        "requirements": ["bar", "bim >= 1"]
    }
    assert definition1.data() == {"identifier": "foo"}


def test_definition_insert_error():
    """Fail to create new definition with non-list element extended."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
    })

    assert definition1.data() == {"identifier": "foo"}

    with pytest.raises(ValueError):
        definition1.insert("identifier", ["test"], 0)


def test_definition_remove():
    """Create new definition from existing definition without element."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
        "version": "0.1.0"
    })

    assert definition1.data() == {
        "identifier": "foo",
        "version": "0.1.0"
    }

    definition2 = definition1.remove("version")
    assert definition2.data() == {"identifier": "foo"}
    assert definition1.data() == {
        "identifier": "foo",
        "version": "0.1.0"
    }


def test_definition_remove_non_existing():
    """Do not raise when removing non existing element."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
    })

    assert definition.data() == {"identifier": "foo"}

    _definition = definition.remove("error")
    assert _definition == definition


def test_definition_remove_key():
    """Create new definition from existing definition without element key."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
        "environ": {
            "key1": "value1",
            "key2": "value2"
        }
    })

    assert definition1.data() == {
        "identifier": "foo",
        "environ": {
            "key1": "value1",
            "key2": "value2"
        }
    }

    definition2 = definition1.remove_key("environ", "key1")
    assert definition2.data() == {
        "identifier": "foo",
        "environ": {
            "key2": "value2"
        }
    }
    assert definition1.data() == {
        "identifier": "foo",
        "environ": {
            "key1": "value1",
            "key2": "value2"
        }
    }


def test_definition_remove_last_key():
    """Create new definition from existing definition without element."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
        "environ": {
            "key1": "value1",
        }
    })

    assert definition1.data() == {
        "identifier": "foo",
        "environ": {
            "key1": "value1",
        }
    }

    definition2 = definition1.remove_key("environ", "key1")
    assert definition2.data() == {
        "identifier": "foo",
    }


def test_definition_remove_key_error():
    """Fail to create new definition without un-existing element or element key.
    """
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
        "environ": {
            "key1": "value1",
        }
    })

    assert definition1.data() == {
        "identifier": "foo",
        "environ": {
            "key1": "value1",
        }
    }

    with pytest.raises(ValueError) as error:
        definition1.remove_key("identifier", "key42")

    assert (
       "Impossible to remove key from 'identifier' as it is not a dictionary."
    ) in str(error)


def test_definition_remove_non_existing_key():
    """Do not raise when removing non existing element key."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "environ": {
            "key1": "value1",
        }
    })

    assert definition.data() == {
        "identifier": "foo",
        "environ": {
            "key1": "value1",
        }
    }

    _definition = definition.remove_key("test", "error")
    assert definition == _definition

    _definition = definition.remove_key("environ", "key42")
    assert definition == _definition


def test_definition_remove_index():
    """Create new definition from existing definition without element index."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
        "requirements": ["test", "bar", "bim >= 1"]
    })

    assert definition1.data() == {
        "identifier": "foo",
        "requirements": ["test", "bar", "bim >= 1"]
    }

    definition2 = definition1.remove_index("requirements", 0)
    assert definition2.data() == {
        "identifier": "foo",
        "requirements": ["bar", "bim >= 1"]
    }
    assert definition1.data() == {
        "identifier": "foo",
        "requirements": ["test", "bar", "bim >= 1"]
    }

    definition3 = definition2.remove_index("requirements", 1)
    assert definition3.data() == {
        "identifier": "foo",
        "requirements": ["bar"]
    }
    assert definition2.data() == {
        "identifier": "foo",
        "requirements": ["bar", "bim >= 1"]
    }
    assert definition1.data() == {
        "identifier": "foo",
        "requirements": ["test", "bar", "bim >= 1"]
    }


def test_definition_remove_last_index():
    """Create new definition from existing definition without element."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
        "requirements": ["test"]
    })

    assert definition1.data() == {
        "identifier": "foo",
        "requirements": ["test"]
    }

    definition2 = definition1.remove_index("requirements", 0)
    assert definition2.data() == {
        "identifier": "foo",
    }


def test_definition_remove_index_error():
    """Fail to create definition without un-existing element or element index.
    """
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
        "requirements": ["test", "bar", "bim >= 1"]
    })

    assert definition1.data() == {
        "identifier": "foo",
        "requirements": ["test", "bar", "bim >= 1"]
    }

    with pytest.raises(ValueError) as error:
        definition1.remove_index("identifier", 42)

    assert (
       "Impossible to remove index from 'identifier' as it is not a list."
    ) in str(error)


def test_definition_remove_non_existing_index():
    """Do not raise when removing non existing element index."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "requirements": ["test"]
    })

    assert definition.data() == {
        "identifier": "foo",
        "requirements": ["test"]
    }

    _definition = definition.remove_index("requirements", 5)
    assert definition == _definition

    _definition = definition.remove_index("test", "error")
    assert definition == _definition


def test_definition_variant_set():
    """Create new definition from existing definition with new variant."""
    definition1 = wiz.definition.Definition({
        "identifier": "foo",
    })

    # Add variant
    definition2 = definition1.set(
        "variants", [
            {
                "identifier": "Variant1",
                "requirements": ["bar"]
            },
        ]
    )
    assert definition2.data() == {
        "identifier": "foo",
        "variants": [
            {
                "identifier": "Variant1",
                "requirements": ["bar"]
            },
        ]
    }
    assert definition1.data() == {"identifier": "foo"}

    # Update variant
    definition3 = definition2.set(
        "variants", [
            {
                "identifier": "Test",
                "requirements": ["bar", "bim > 1"]
            }
        ]
    )

    assert definition3.data() == {
        "identifier": "foo",
        "variants": [
            {
                "identifier": "Test",
                "requirements": ["bar", "bim > 1"]
            }
        ]
    }
    assert definition2.data() == {
        "identifier": "foo",
        "variants": [
            {
                "identifier": "Variant1",
                "requirements": ["bar"]
            }
        ]
    }
    assert definition1.data() == {"identifier": "foo"}


def test_definition_non_mutated_input():
    """Ensure that input mapping isn't mutated when creating definition."""
    data = {
        "identifier": "test",
        "version": "0.1.0",
        "namespace": "foo",
        "description": "This is a definition",
        "auto-use": True,
        "system": {
            "platform": "linux",
            "os": "el >= 6, < 7",
            "arch": "x86_64"
        },
        "command": {
            "app": "AppX"
        },
        "environ": {
            "KEY1": "VALUE1"
        },
        "requirements": ["foo"],
        "conditions": ["baz"],
        "variants": [
            {
                "identifier": "Variant1",
                "command": {
                    "appV1": "AppX --test"
                },
                "environ": {
                    "KEY2": "VALUE2"
                },
                "requirements": ["bim >= 9, < 10"]
            }
        ]
    }

    _data = copy.deepcopy(data)
    wiz.definition.Definition(_data)

    assert data == _data



# :coding: utf-8

import itertools
import os
import re

import wiz.config

#: Compiled regular expression to identify environment variables in string.
ENV_PATTERN = re.compile(r"\${(\w+)}|\$(\w+)")


def initiate(mapping=None):
    """Return the minimal environment mapping to augment.

    :param mapping: Custom environment mapping which will be added to the
        initial environment. Default is None.

    .. seealso:: :ref:`configuration/initial_environment`

    """
    config = wiz.config.fetch()
    environ = config.get("environ", {}).get("initial", {})

    for key in config.get("environ", {}).get("passthrough", []):
        value = os.environ.get(key)

        if value:
            environ[key] = value

    if mapping is not None:
        environ.update(**mapping)

    return environ


def sanitize(mapping):
    """Return sanitized environment *mapping*.

    Resolve all key references within *mapping* values and remove all
    self-references::

        >>> sanitize({
        ...     "PLUGIN": "${HOME}/.app:/path/to/somewhere:${PLUGIN}",
        ...     "HOME": "/usr/people/me"
        ... })

        {
            "HOME": "/usr/people/me",
            "PLUGIN": "/usr/people/me/.app:/path/to/somewhere"
        }

    :param mapping: Environment mapping to sanitize.

    :return: Sanitized environment mapping.

    """
    _mapping = {}

    # Remove all self reference environment variables in values with trailing
    # path separator (e.g. {"PATH": "/path/to/somewhere:${PATH}"}).
    pattern = r"(\${{{0}}}:?|:?\${{{0}}}|\${0}:?|:?\${0})"

    for key, value in mapping.items():
        _mapping[key] = re.sub(pattern.format(key), lambda _: "", value)

    # Last pass across mapping to substitute remaining variables.
    for key, value in _mapping.items():
        _mapping[key] = substitute(value, _mapping)

    return _mapping


def contains(text, name):
    """Indicate whether *text* contains a reference to variable *name*.

    Example::

        >>> contains("${HOME}/path/to/data", "HOME")

        True

    :param text: String which can contain environment variable
        (e.g. "${PATH}/to/somewhere").

    :param name: Name of an environment variable (e.g. "PATH").

    :return: Boolean value.

    """
    return name in itertools.chain(*ENV_PATTERN.findall(text))


def substitute(text, environment):
    """Substitute all environment variables in *text* from *environment*.

    Example::

        >>> substitute("${HOME}/path/to/data", {"HOME": "/usr/people/john-doe"})

        /usr/people/john-doe/path/to/data

    :param text: String which can contain environment variable
        (e.g. "${PATH}/to/somewhere").

    :param environment: Mapping of environment variables with their respective
        values.

    :return: Resolved text string.

    """

    def _substitute(match):
        origin = match.group(0)
        name = next(item for item in match.groups() if item is not None)
        return environment.get(name, origin)

    return ENV_PATTERN.sub(_substitute, text)



# :coding: utf-8

import base64
import collections
import copy
import hashlib
import pipes
import re
import zlib

import colorama
from packaging.requirements import InvalidRequirement
from packaging.version import Version, InvalidVersion
import ujson

import wiz.exception
import wiz.symbol
from ._requirement import Requirement

# Arbitrary number which indicates a very high version number
_INFINITY_VERSION = 9999


def get_requirement(content):
    """Return the corresponding requirement instance from *content*.

    :param content: String representing a requirement, with or without
        version specifier or variant (e.g. "maya", "nuke >= 10, < 11",
        "ldpk-nuke[10.0]").

    :return: Instance of :class:`packaging.requirements.Requirement`.

    :raise: :exc:`wiz.exception.InvalidRequirement` if the requirement is
        incorrect.

    """
    try:
        return Requirement(content)
    except InvalidRequirement:
        raise wiz.exception.RequirementError(
            "The requirement '{}' is incorrect".format(content)
        )


def get_requirements(contents):
    """Return the corresponding requirement instance from *content*.

    :param contents: List of strings representing requirements, with or without
        version specifier or variant (e.g. "maya", "nuke >= 10, < 11",
        "ldpk-nuke[10.0]").

    :return: List of :class:`packaging.requirements.Requirement` instances.

    :raise: :exc:`wiz.exception.InvalidRequirement` if the requirement is
        incorrect.

    """
    return [get_requirement(content) for content in contents]


def get_version(content):
    """Return the corresponding version instance from *content*.

    :param content: String representing a version (e.g. "2018", "0.1.0").

    :return: Instance of :class:`packaging.version.Version`.

    :raise: :exc:`wiz.exception.InvalidVersion` if the version is incorrect.

    """
    try:
        return Version(content)
    except InvalidVersion:
        raise wiz.exception.VersionError(
            "The version '{}' is incorrect".format(content)
        )


def is_overlapping(requirement1, requirement2):
    """Indicate whether requirements are overlapping.

    A requirement is overlapping with another one if their intersection of
    version ranges are not empty.

    Example::

        >>> is_overlapping(Requirement("foo >= 10"), Requirement("foo < 9"))
        True

        >>> is_overlapping(Requirement("foo >= 10"), Requirement("foo < 8"))
        False

        >>> is_overlapping(Requirement("foo[V2]"), Requirement("foo[V1]"))
        False

    :param requirement1: Instance of
        :class:`packaging.requirements.Requirement`.

    :param requirement2: Instance of
        :class:`packaging.requirements.Requirement`.

    :return: Boolean value.

    :raise: :exc:`ValueError` if requirements cannot be compared.

    """
    if requirement1.name != requirement2.name:
        raise ValueError(
            "Impossible to compare requirements with different names "
            "['{}' and '{}'].".format(requirement1.name, requirement2.name)
        )

    if requirement1.extras != requirement2.extras:
        return False

    r1 = extract_version_ranges(requirement1)
    r2 = extract_version_ranges(requirement2)
    return (
        (r2[-1][1] is None or r1[0][0] is None or r2[-1][1] >= r1[0][0]) and
        (r1[-1][1] is None or r2[0][0] is None or r1[-1][1] >= r2[0][0])
    )


def extract_version_ranges(requirement):
    """Extract version ranges from *requirement*.

    Requirement could contain the following specifiers:

    * `Compatible release
      <https://www.python.org/dev/peps/pep-0440/#compatible-release>`_
    * `Version matching
      <https://www.python.org/dev/peps/pep-0440/#version-matching>`_
    * `Version exclusion
      <https://www.python.org/dev/peps/pep-0440/#version-exclusion>`_
    * `Inclusive ordered comparison
      <https://www.python.org/dev/peps/pep-0440/#inclusive-ordered-comparison>`_
    * `Exclusive ordered comparison
      <https://www.python.org/dev/peps/pep-0440/#exclusive-ordered-comparison>`_

    Example::

        >>> extract_version_ranges(Requirement("foo"))
        [(None, None)]

        >>> extract_version_ranges(Requirement("foo==2.1.1"))
        [((2, 1, 1), (2, 1, 1))]

        >>> extract_version_ranges(Requirement("foo!=2.1.1"))
        [(None, (2, 1, 0, 9999)), ((2, 1, 1, 1), None)]

        >>> extract_version_ranges(Requirement("foo ==2.*"))
        [((2,), (2, 9999))]

        >>> extract_version_ranges(Requirement("foo >= 2, < 3"))
        [((2,), (2, 9999))]

    :param requirement: Instance of :class:`packaging.requirements.Requirement`.

    :return: List of version tuples.

    :raise: :exc:`wiz.exception.InvalidVersion` if the version extracted from
        the specifier is incorrect.

    :raise: :exc:`wiz.exception.InvalidRequirement` if the specifier operator
        is not accepted or if the requirement does not allow any versions to be
        reached.

    """
    version_ranges = [(None, None)]

    # Sort specifiers for deterministic results.
    for specifier in sorted(
        requirement.specifier, key=lambda s: (s.version, s.operator)
    ):
        # Extract version number.
        _version = wiz.utility.get_version(
            re.sub(r"\.\*$", "", specifier.version)
        ).release

        # Update maximum according to operator.
        if specifier.operator == ">=":
            _update_minimum_version(_version, version_ranges)

        elif specifier.operator == "<=":
            _update_maximum_version(_version, version_ranges)

        elif specifier.operator == ">":
            _min_version = _increment_version(_version)
            _update_minimum_version(_min_version, version_ranges)

        elif specifier.operator == "<":
            _max_version = _decrement_version(_version)
            _update_maximum_version(_max_version, version_ranges)

        elif specifier.operator == "~=":
            _max_version = _increment_version(_version, delta=_INFINITY_VERSION)
            _update_minimum_version(_version, version_ranges)
            _update_maximum_version(_max_version, version_ranges)

        elif specifier.operator == "==" and specifier.version.endswith(".*"):
            _max_version = _increment_version(_version, delta=_INFINITY_VERSION)
            _update_minimum_version(_version, version_ranges)
            _update_maximum_version(_max_version, version_ranges)

        elif specifier.operator == "==":
            _update_minimum_version(_version, version_ranges)
            _update_maximum_version(_version, version_ranges)

        elif specifier.operator == "!=" and specifier.version.endswith(".*"):
            _min_version = _decrement_version(_version)
            _max_version = _increment_version(_version, add_subversion=False)
            _update_version_ranges((_min_version, _max_version), version_ranges)

        elif specifier.operator == "!=":
            _min_version = _decrement_version(_version)
            _max_version = _increment_version(_version)
            _update_version_ranges((_min_version, _max_version), version_ranges)

        else:
            raise wiz.exception.RequirementError(
                "Operator '{}' is not accepted for requirement '{}'".format(
                    specifier.operator, requirement
                )
            )

    return version_ranges


def _update_maximum_version(version, ranges):
    """Update version *ranges* with maximum *version*.

    Example::

        >>> version_ranges = [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 4))]
        >>> _update_maximum_version((1, 2, 3), version_ranges)
        >>> print(version_ranges)
        [((1, 2, 3), (1, 2, 3))]

    :param version: Tuple representing a version (e.g. (1,2,3)).

    :param ranges: ordered list of tuples containing two ordered version release
        tuples (e.g [((1,2,3), (1,3,0)), (1,3,3), (1,4))]).

    .. warning::

        The *ranges* list will be mutated.

    """
    _ranges = []

    if ranges[0][0] is not None and version < ranges[0][0]:
        raise wiz.exception.RequirementError(
            "The requirement is incorrect as maximum value '{}' cannot be set "
            "when minimum value is '{}'.".format(
                ".".join(str(v) for v in version),
                ".".join(str(v) for v in ranges[0][0])
            )
        )

    for start_version, end_version in ranges:
        if start_version and start_version > version:
            break
        if not end_version or end_version >= version:
            _ranges.append((start_version, version))
            break
        _ranges.append((start_version, end_version))

    # Mutated input ranges
    ranges[:] = _ranges


def _update_minimum_version(version, ranges):
    """Update version *ranges* with minimum *version*.

    Example::

        >>> version_ranges = [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 4))]
        >>> _update_minimum_version((1, 3, 3), version_ranges)
        >>> print(version_ranges)
        [((1, 3, 3), (1, 4))]

    :param version: Tuple representing a version (e.g. (1,2,3)).

    :param ranges: ordered list of tuples containing two ordered version release
        tuples (e.g [((1,2,3), (1,3,0)), (1,3,3), (1,4))]).

    .. warning::

        The *ranges* list will be mutated.

    """
    _ranges = []

    if ranges[-1][1] is not None and version > ranges[-1][1]:
        raise wiz.exception.RequirementError(
            "The requirement is incorrect as minimum value '{}' cannot be set "
            "when maximum value is '{}'.".format(
                ".".join(str(v) for v in version),
                ".".join(str(v) for v in ranges[-1][1])
            )
        )

    for start_version, end_version in reversed(ranges):
        if end_version and end_version < version:
            break
        if not start_version or start_version <= version:
            _ranges = [(version, end_version)] + _ranges
            break
        _ranges = [(start_version, end_version)] + _ranges

    # Mutated input ranges
    ranges[:] = _ranges


def _update_version_ranges(excluded, ranges):
    """Update version *ranges* from *excluded* version range.

    Example::

        >>> version_ranges = [((1,2,3), (1,3,0)), ((1,3,3), (1,4))]
        >>> _update_version_ranges(((1,2,3), (1,3,3)), version_ranges)
        >>> print(version_ranges)
        [((1, 2, 3), (1, 2, 3)), ((1, 3, 3), (1, 4))]

    :param excluded: Tuple containing two ordered version release tuples (e.g.
        ((1,2,3), (1,3,0))). These two versions are included in *ranges*, but
        all versions in between should be excluded.

    :param ranges: ordered list of tuples containing two ordered version release
        tuples (e.g [((1,2,3), (1,3,0)), (1,3,3), (1,4))]).

    .. warning::

        The *ranges* list will be mutated.

    """
    _ranges = []

    for r in ranges:
        out_before = r[0] is not None and r[0] >= excluded[1]
        out_after = r[1] is not None and r[1] <= excluded[0]

        # Exclusion zone is outside of range.
        if out_before or out_after:
            _ranges.append(r)
            continue

        r0_excluded = r[0] is not None and r[0] > excluded[0]
        r1_excluded = r[1] is not None and r[1] < excluded[1]

        # Exclusion zone covers all range.
        if r0_excluded and r1_excluded:
            continue

        # Exclusion zone cover start of range only.
        elif r0_excluded and not r1_excluded:
            _ranges.append((excluded[1], r[1]))

        # Exclusion zone cover end of range only.
        elif not r0_excluded and r1_excluded:
            _ranges.append((r[0], excluded[0]))

        # Exclusion zone cover middle of range.
        elif not r0_excluded and not r1_excluded:
            _ranges.append((r[0], excluded[0]))
            _ranges.append((excluded[1], r[1]))

    if len(_ranges) == 0:
        raise wiz.exception.RequirementError(
            "The requirement is incorrect as excluded version range '{}-{}' "
            "makes all other versions unreachable.".format(
                ".".join(str(v) for v in excluded[0]),
                ".".join(str(v) for v in excluded[1]),
            )
        )

    # Mutated input ranges
    ranges[:] = _ranges


def _increment_version(version, delta=1, add_subversion=True):
    """Increment *version*.

    This will attempt to increase to the nearest possible version tuple.

    Example::

        >>> _increment_version((1, 2, 0))
        (1, 2, 0, 1)

        >>> _increment_version((1, 2, 0), add_subversion=False)
        (1, 2, 1)

        >>> _increment_version((1, 1, 1), delta=3)
        (1, 1, 1, 3)

    :param version: Tuple representing a version (e.g. (1,2,3)).

    :param delta: Number to add to the minimal the *version*. Default is 1.

    :param add_subversion: Indicate whether a sub-version should be used instead
        of only increasing the minimal release version. Default is True.

    :return: New version tuple.

    """
    if not add_subversion:
        return version[:-1] + (version[-1] + delta,)
    return version + (delta,)


def _decrement_version(version):
    """Decrement *version*.

    This will attempt to decrease to the nearest possible version tuple.

    Example::

        >>> _decrement_version((1,))
        (0, 9999)

        >>> _decrement_version((1, 0, 0))
        (1, 9999)

        >>> _decrement_version((1, 2, 0))
        (1, 1, 9999)

        >>> _decrement_version((1, 1, 1))
        (1, 1, 0, 9999)

    :param version: Tuple representing a version (e.g. (1,2,3)).

    :return: New version tuple.

    """
    index = -1
    while version[index] == 0:
        index -= 1

    return version[:index] + (version[index] - 1, _INFINITY_VERSION)


def encode(element):
    """Return serialized and encoded *element*.

    *element* is serialized first, then encoded into :term:`base64`.

    :param element: Content to encode.

    :return: Encoded string.

    :raise: :exc:`TypeError` if *element* is not JSON serializable.

    """
    serialized = ujson.dumps(element).encode("utf-8")
    return base64.b64encode(zlib.compress(serialized)).decode("utf-8")


def decode(element):
    """Return deserialized and decoded *element*.

    *element* is decoded first from :term:`base64`, then deserialized.

    :param element: Content to decode.

    :raise: :exc:`TypeError` if *element* cannot be decoded or deserialized.

    """
    return ujson.loads(zlib.decompress(base64.b64decode(element)))


def compute_label(definition):
    """Return unique label for *definition*.

    The name should be in the form of::

        "'foo'"
        "'bar' [0.1.0]"
        "'baz' [0.2.0] (linux : el =! 7)"
        "'bim' (linux : el >= 6, < 7)"

    :param definition: Instance of :class:`wiz.definition.Definition`.

    :return: String representing definition.

    """
    label = "'{}'".format(definition.qualified_identifier)

    if definition.version:
        label += " [{}]".format(definition.version)

    if definition.system:
        system_identifier = compute_system_label(definition)
        label += " ({})".format(system_identifier)

    return label


def compute_system_label(definition):
    """Return unique system label from *definition*.

    The system identifier should be in the form of::

        "noarch"
        "linux : x86_64 : el >= 7, < 8"
        "centos >= 7, < 8"
        "x86_64 : el >= 7, < 8"
        "windows"

    :param definition: Instance of :class:`wiz.definition.Definition`.

    :return: String representing system identifier.

    """
    elements = [
        definition.system.get(element)
        for element in ["platform", "arch", "os"]
        if definition.system.get(element) is not None
    ]
    return " : ".join(elements) or "noarch"


def compute_file_name(definition):
    """Return unique file name from *definition*.

    The file name should be in the form of::

        "foo.json"
        "namespace-foo-0.1.0.json"
        "foo-0.1.0.json"
        "foo-0.1.0-M2Uq9Esezm-m00VeWkTzkQIu3T4.json"

    :param definition: Instance of :class:`wiz.definition.Definition`.

    :return: File name representing definition.

    """
    name = definition.identifier

    if definition.namespace:
        namespace = "-".join(
            definition.namespace.split(wiz.symbol.NAMESPACE_SEPARATOR)
        )
        name = "{}-{}".format(namespace, name)

    if definition.version:
        name += "-{}".format(definition.version)

    if definition.system:
        system_identifier = wiz.utility.compute_system_label(definition)
        data = re.sub(r"(\s+|:+)", "", system_identifier).encode("utf-8")
        encoded = base64.urlsafe_b64encode(hashlib.sha1(data).digest())
        name += "-{}".format(encoded.rstrip(b"=").decode("utf-8"))

    return "{}.json".format(name)


def combine_command(elements):
    """Return command *elements* as a string.

    Example::

        >>> combine_command(
        ...     ['python2.7', '-c', 'import os; print(os.environ["HOME"])'])
        ... )

        python2.7 -c 'import os; print(os.environ["HOME"])'

    :param elements: List of strings constituting the command line to execute
        (e.g. ["app_exe", "--option", "value"])


    """
    return " ".join([pipes.quote(element) for element in elements])


def colored_text(message, color):
    """Return colored *message* according to color *name*.

    Available color names are: "black", "red", "green", "yellow", "blue",
    "magenta", "cyan" and "white".

    :param message: String to colorize.

    :param color: Symbol of color to apply to *message*.

    :return: Colorized message.

    """
    return (
        getattr(colorama.Fore, color.upper()) + message +
        colorama.Style.RESET_ALL
    )


def deep_update(mapping1, mapping2):
    """Recursively update *mapping1* from *mapping2*.

    Contrary to :meth:`dict.update`, this function will attempt to update
    sub-dictionaries defined in both mappings instead of overwriting the value
    defined in *mapping1*::

        >>> deep_update({"A": {"B": 2}}, {"A": {"C": 3}})
        {"A": {"B": 2, "C": 3}}

    :param mapping1: Mapping to update

    :param mapping2: Mapping to update *mapping1* from

    :return: *mapping1* mutated.

    .. note::

        *mapping1* will be mutated, but *mapping2* will not.

    """
    for key, value in mapping2.items():
        if isinstance(value, collections.Mapping):
            mapping1[key] = deep_update(mapping1.get(key, {}), value)
        else:
            mapping1[key] = value
    return mapping1


def sanitize_requirement(requirement, package):
    """Return qualified *requirement* depending on *package*'s namespace.

    Example::

        # If the package has a namespace of 'foo'
        >>> sanitize_requirement(Requirement("A >=1, <2"), package)
        Requirement("foo::A >=1, <2")

        # If the package has no namespace
        >>> sanitize_requirement(Requirement("A >=1, <2"), package)
        Requirement("::A >=1, <2")

    :param requirement: Instance of :class:`packaging.requirements.Requirement`.

    :param package: Instance of :class:`wiz.package.Package`.

    :return: new instance of :class:`packaging.requirements.Requirement`.

    :raise: :exc:`ValueError` if *requirement* is incompatible with *package*.

    """
    if not match(requirement, package):
        raise ValueError(
            "Requirement '{0}' is incompatible with package '{1}'".format(
                requirement, package.identifier
            )
        )

    # Prevent mutating incoming requirement.
    _requirement = copy.deepcopy(requirement)
    _requirement.name = wiz.symbol.NAMESPACE_SEPARATOR.join([
        package.namespace or "", package.definition.identifier
    ])
    return _requirement


def compute_namespace_counter(requirements, definition_mapping):
    """Compute namespace frequency counter from *requirements*.

    :param requirements: List of :class:`packaging.requirements.Requirement`
        instances.

    :param definition_mapping: Mapping regrouping all available definitions
        associated with their unique identifier.

    :return: Instance of :class:`collections.Counter`.

    """
    mapping = definition_mapping.get("__namespace__", {})

    namespaces = []

    for requirement in requirements:
        namespace, _ = wiz.utility.extract_namespace(requirement)
        if namespace is not None:
            namespaces.append(namespace)
        else:
            namespaces += mapping.get(requirement.name, [])

    return collections.Counter(namespaces)


def match(requirement, package):
    """Return whether *requirement* is compatible with *package*.

    :param requirement: Instance of :class:`packaging.requirements.Requirement`.

    :param package: Instance of :class:`wiz.package.Package`.

    :return: Boolean value.

    """
    namespace, identifier = extract_namespace(requirement)

    # Ignore if package identifier doesn't match requirement name.
    if package.definition.identifier != identifier:
        return False

    # Ignore if package namespace doesn't match requirement name.
    if namespace is not None and package.definition.namespace != namespace:
        return False

    # Ignore if package variant doesn't match any requirement extras.
    variants_requested = list(requirement.extras)
    if len(variants_requested) > 0:
        variant_identifier = package.variant_identifier
        if not any(_id == variant_identifier for _id in variants_requested):
            return False

    # Node is matching if package has no version.
    if package.version is not None:
        # Remove namespace from requirement so specifier can be used.
        _requirement = copy.deepcopy(requirement)
        _requirement.name = identifier

        if not _requirement.specifier.contains(package.version):
            return False

    return True


def extract_namespace(requirement):
    """Extract namespace and identifier from *requirement*.

    Example::

        >>> extract_namespace(Requirement("foo"))
        None, "foo"

        >>> extract_namespace(Requirement("::foo"))
        None, "foo"

        >>> extract_namespace(Requirement("bar::foo"))
        "bar", "foo"

        >>> extract_namespace(Requirement("bar1::bar2::foo"))
        "bar1::bar2", "foo"

    """
    try:
        namespace, identifier = requirement.name.rsplit(
            wiz.symbol.NAMESPACE_SEPARATOR, 1
        )

        if not len(namespace):
            namespace = None

    except ValueError:
        identifier = requirement.name
        namespace = None

    return namespace, identifier


def check_conflicting_requirements(package1, package2):
    """Check whether some requirements are conflicting between packages.

    :param package1: Instance of :class:`wiz.package.Package`.

    :param package2: Instance of :class:`wiz.package.Package` to compare
        *package1* with.

    :return: Boolean value.

    """
    mapping = {}

    for requirement in package1.requirements + package2.requirements:
        _requirement = mapping.get(requirement.name)
        if _requirement is not None:
            if not is_overlapping(requirement, _requirement):
                return True

        mapping[requirement.name] = requirement

    return False



# :coding: utf-8

import os
import os.path
import types

import pytest

import wiz.config
import wiz.exception
import wiz.filesystem
import wiz.registry


@pytest.fixture(autouse=True)
def reset_configuration(mocker):
    """Ensure that no personal configuration is fetched during tests."""
    mocker.patch.object(os.path, "expanduser", return_value="__HOME__")

    # Reset configuration.
    wiz.config.fetch(refresh=True)


@pytest.fixture()
def mocked_remove(mocker):
    """Return mocked 'os.remove' function."""
    return mocker.patch.object(os, "remove")


@pytest.fixture()
def mocked_export_definition(mocker):
    """Return mocked 'wiz.export_definition' function."""
    return mocker.patch.object(wiz, "export_definition")


@pytest.fixture()
def mocked_fetch_definition_mapping(mocker):
    """Return mocked 'wiz.fetch_definition_mapping' function."""
    return mocker.patch.object(wiz, "fetch_definition_mapping")


@pytest.fixture()
def mocked_fetch_definition(mocker):
    """Return mocked 'wiz.fetch_definition' function."""
    return mocker.patch.object(wiz, "fetch_definition")


@pytest.fixture()
def mocked_filesystem_accessible(mocker):
    """Return mocked 'wiz.filesystem.is_accessible' function."""
    return mocker.patch.object(wiz.filesystem, "is_accessible")


@pytest.fixture()
def mocked_filesystem_get_name(mocker):
    """Return mocked 'wiz.filesystem.get_name' function."""
    return mocker.patch.object(wiz.filesystem, "get_name")


@pytest.fixture()
def mocked_user_home(mocker, temporary_directory):
    """Return mocked local home path."""
    mocker.patch.object(
        os.path, "expanduser", return_value=temporary_directory
    )
    return temporary_directory


@pytest.fixture()
def mocked_local(mocker):
    """Return mocked local registry path."""
    return mocker.patch.object(
        wiz.registry, "get_local", return_value="/usr/people/me/.wiz/registry"
    )


@pytest.fixture()
def mocked_discover(mocker):
    """Return mocked working directory registry paths."""
    paths = [
        "/jobs/ads/project/.common/wiz/registry",
        "/jobs/ads/project/identity/shot/.common/wiz/registry",
    ]
    return mocker.patch.object(
        wiz.registry, "discover", return_value=(path for path in paths)
    )


@pytest.fixture()
def mocked_definitions():
    """Return mocked simple definition list."""
    return [
        wiz.definition.Definition({
            "identifier": "foo",
            "version": "0.1.0",
        }),
        wiz.definition.Definition({
            "identifier": "bar",
        }),
        wiz.definition.Definition({
            "identifier": "baz",
            "version": "2.5.1",
        }),
    ]


@pytest.mark.usefixtures("mocked_user_home")
def test_get_local_unreachable():
    """Return local registry."""
    assert wiz.registry.get_local() is None


def test_get_defaults():
    """Return default registries."""
    assert wiz.registry.get_defaults() == []


def test_discover(mocked_filesystem_accessible):
    """Discover registries under paths."""
    mocked_filesystem_accessible.side_effect = [False, True, False, True]

    path = os.path.join(os.sep, "project", "identity", "shot", "animation")
    registries = wiz.registry.discover(path)

    end = os.path.join(".wiz", "registry")

    assert isinstance(registries, types.GeneratorType)
    assert list(registries) == [
        os.path.join(os.sep, "project", "identity", end),
        os.path.join(os.sep, "project", "identity", "shot", "animation", end)
    ]

    assert mocked_filesystem_accessible.call_count == 4
    mocked_filesystem_accessible.assert_any_call(
        os.path.join(os.sep, "project", end),
    )
    mocked_filesystem_accessible.assert_any_call(
        os.path.join(os.sep, "project", "identity", end),
    )
    mocked_filesystem_accessible.assert_any_call(
        os.path.join(os.sep, "project", "identity", "shot", end),
    )
    mocked_filesystem_accessible.assert_any_call(
        os.path.join(os.sep, "project", "identity", "shot", "animation", end),
    )


def test_discover_with_prefix(mocked_filesystem_accessible):
    """Discover registries under prefix."""
    mocked_filesystem_accessible.side_effect = [False, True, False, True]

    # Add prefix in config.
    prefix = os.path.join(os.sep, "jobs", "ads")
    config = wiz.config.fetch()
    config["registry"]["discovery_prefix"] = prefix

    path = os.path.join(prefix, "project", "identity", "shot", "animation")
    registries = wiz.registry.discover(path)

    end = os.path.join(".wiz", "registry")

    assert isinstance(registries, types.GeneratorType)
    assert list(registries) == [
        os.path.join(prefix, "project", "identity", end),
        os.path.join(prefix, "project", "identity", "shot", "animation", end)
    ]

    assert mocked_filesystem_accessible.call_count == 4
    mocked_filesystem_accessible.assert_any_call(
        os.path.join(prefix, "project", end),
    )
    mocked_filesystem_accessible.assert_any_call(
        os.path.join(prefix, "project", "identity", end),
    )
    mocked_filesystem_accessible.assert_any_call(
        os.path.join(prefix, "project", "identity", "shot", end),
    )
    mocked_filesystem_accessible.assert_any_call(
        os.path.join(prefix, "project", "identity", "shot", "animation", end),
    )


def test_discover_with_prefix_fail(mocked_filesystem_accessible):
    """Fail to discover registries under paths not in prefix"""
    mocked_filesystem_accessible.side_effect = [False, True, False, True]

    # Add prefix in config.
    prefix = os.path.join(os.sep, "jobs", "ads")
    config = wiz.config.fetch()
    config["registry"]["discovery_prefix"] = prefix

    path = os.path.join(
        os.sep, "somewhere", "else", "project", "identity", "shot", "animation"
    )
    registries = wiz.registry.discover(path)

    assert isinstance(registries, types.GeneratorType)
    assert list(registries) == []

    mocked_filesystem_accessible.assert_not_called()


@pytest.mark.parametrize("options, paths, expected", [
    (
        {},
        ["/path/to/registry1", "path/to/registry2"],
        [
            "/path/to/registry1",
            os.path.join(os.getcwd(), "path/to/registry2"),
            "/jobs/ads/project/.common/wiz/registry",
            "/jobs/ads/project/identity/shot/.common/wiz/registry",
            "/usr/people/me/.wiz/registry"
        ]
    ),
    (
        {"include_local": False},
        ["/path/to/registry1", "path/to/registry2"],
        [
            "/path/to/registry1",
            os.path.join(os.getcwd(), "path/to/registry2"),
            "/jobs/ads/project/.common/wiz/registry",
            "/jobs/ads/project/identity/shot/.common/wiz/registry"
        ]
    ),
    (
        {"include_working_directory": False},
        ["/path/to/registry1", "path/to/registry2"],
        [
            "/path/to/registry1",
            os.path.join(os.getcwd(), "path/to/registry2"),
            "/usr/people/me/.wiz/registry"
        ]
    ),
    (
        {"include_local": False, "include_working_directory": False},
        ["/path/to/registry1", os.path.join(os.getcwd(), "path/to/registry2")],
        ["/path/to/registry1", os.path.join(os.getcwd(), "path/to/registry2")]
    )
], ids=[
    "default",
    "without-local",
    "without-cwd",
    "without-local-nor-cwd",
])
@pytest.mark.usefixtures("mocked_local")
@pytest.mark.usefixtures("mocked_discover")
def test_fetch(mocked_filesystem_accessible, options, paths, expected):
    """Fetch the registries."""
    mocked_filesystem_accessible.return_value = True
    assert wiz.registry.fetch(paths, **options) == expected


@pytest.mark.usefixtures("mocked_discover")
def test_fetch_unreachable_local(mocked_filesystem_accessible, mocked_local):
    mocked_filesystem_accessible.return_value = True
    mocked_local.return_value = None

    paths = ["/path/to/registry1", "path/to/registry2"]
    assert wiz.registry.fetch(paths) == [
        "/path/to/registry1",
        os.path.join(os.getcwd(), "path/to/registry2"),
        "/jobs/ads/project/.common/wiz/registry",
        "/jobs/ads/project/identity/shot/.common/wiz/registry"
    ]


@pytest.mark.usefixtures("mocked_discover")
@pytest.mark.usefixtures("mocked_local")
def test_fetch_unreachable_paths(mocked_filesystem_accessible):
    mocked_filesystem_accessible.side_effect = [True, False]

    paths = ["/path/to/registry1", "path/to/registry2"]
    assert wiz.registry.fetch(paths) == [
        "/path/to/registry1",
        "/jobs/ads/project/.common/wiz/registry",
        "/jobs/ads/project/identity/shot/.common/wiz/registry",
        "/usr/people/me/.wiz/registry"
    ]


def test_install_to_path(
    temporary_directory, mocked_definitions, mocked_fetch_definition_mapping,
    mocked_fetch_definition, mocked_export_definition, mocked_remove, logger,
):
    """Install definitions to path."""
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_fetch_definition.side_effect = wiz.exception.RequestNotFound("Error")

    wiz.registry.install_to_path(
        mocked_definitions, temporary_directory
    )

    mocked_fetch_definition_mapping.assert_called_once_with(
        [temporary_directory]
    )

    assert mocked_fetch_definition.call_count == 3
    mocked_fetch_definition.assert_any_call("foo==0.1.0", "__MAPPING__")
    mocked_fetch_definition.assert_any_call("bar", "__MAPPING__")
    mocked_fetch_definition.assert_any_call("baz==2.5.1", "__MAPPING__")

    mocked_remove.assert_not_called()

    assert mocked_export_definition.call_count == 3
    mocked_export_definition.assert_any_call(
        temporary_directory, mocked_definitions[0], overwrite=True
    )
    mocked_export_definition.assert_any_call(
        temporary_directory, mocked_definitions[1], overwrite=True
    )
    mocked_export_definition.assert_any_call(
        temporary_directory, mocked_definitions[2], overwrite=True
    )

    logger.info.assert_called_once_with(
        "Successfully installed 3 definition(s) to registry '{}'."
        .format(temporary_directory)
    )


def test_install_to_path_with_relative_path(
    temporary_directory, mocked_definitions, mocked_fetch_definition_mapping,
    mocked_fetch_definition, mocked_export_definition, mocked_remove, logger,
):
    """Install definitions to relative registry path."""
    path = os.path.join(temporary_directory, "path", "..", "foo")
    os.makedirs(path)

    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_fetch_definition.side_effect = wiz.exception.RequestNotFound("Error")

    wiz.registry.install_to_path(
        mocked_definitions, path
    )

    registry_path = os.path.join(temporary_directory, "foo")
    mocked_fetch_definition_mapping.assert_called_once_with(
        [registry_path]
    )

    assert mocked_fetch_definition.call_count == 3
    mocked_fetch_definition.assert_any_call("foo==0.1.0", "__MAPPING__")
    mocked_fetch_definition.assert_any_call("bar", "__MAPPING__")
    mocked_fetch_definition.assert_any_call("baz==2.5.1", "__MAPPING__")

    mocked_remove.assert_not_called()

    assert mocked_export_definition.call_count == 3
    mocked_export_definition.assert_any_call(
        registry_path, mocked_definitions[0], overwrite=True
    )
    mocked_export_definition.assert_any_call(
        registry_path, mocked_definitions[1], overwrite=True
    )
    mocked_export_definition.assert_any_call(
        registry_path, mocked_definitions[2], overwrite=True
    )

    logger.info.assert_called_once_with(
        "Successfully installed 3 definition(s) to registry '{}'."
        .format(registry_path)
    )


def test_install_to_path_error_path(
    temporary_directory, mocked_definitions, mocked_fetch_definition_mapping,
    mocked_fetch_definition, mocked_export_definition, mocked_remove, logger,
):
    """Fail to install definitions when path is incorrect."""
    registry_path = os.path.join(temporary_directory, "somewhere")

    with pytest.raises(wiz.exception.InstallError) as error:
        wiz.registry.install_to_path(mocked_definitions, registry_path)

    mocked_remove.assert_not_called()
    mocked_fetch_definition_mapping.assert_not_called()
    mocked_fetch_definition.assert_not_called()
    mocked_export_definition.assert_not_called()
    logger.info.assert_not_called()

    assert (
        "{!r} is not a valid registry directory.".format(registry_path)
    ) in str(error)


def test_install_to_path_error_definition_exists(
    temporary_directory, mocked_definitions, mocked_fetch_definition_mapping,
    mocked_fetch_definition, mocked_export_definition, mocked_remove, logger,
):
    """Fail to install definitions when definition exists."""
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_fetch_definition.side_effect = [
        wiz.exception.RequestNotFound("Error"),
        wiz.definition.Definition({
            "identifier": "bar",
        }),
        wiz.definition.Definition({
            "identifier": "baz",
            "version": "2.5.1",
            "description": "test",
        }),
    ]

    with pytest.raises(wiz.exception.DefinitionsExist) as error:
        wiz.registry.install_to_path(mocked_definitions, temporary_directory)

    mocked_fetch_definition_mapping.assert_called_once_with(
        [temporary_directory]
    )

    assert mocked_fetch_definition.call_count == 3
    mocked_fetch_definition.assert_any_call("foo==0.1.0", "__MAPPING__")
    mocked_fetch_definition.assert_any_call("bar", "__MAPPING__")
    mocked_fetch_definition.assert_any_call("baz==2.5.1", "__MAPPING__")

    mocked_remove.assert_not_called()
    mocked_export_definition.assert_not_called()
    logger.info.assert_not_called()

    assert (
        "DefinitionsExist: 1 definition(s) already exist in registry."
    ) in str(error)


def test_install_to_path_overwrite(
    temporary_directory, mocked_definitions, mocked_fetch_definition_mapping,
    mocked_fetch_definition, mocked_export_definition, mocked_remove, logger,
):
    """Install definitions while overwriting existing definitions."""
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_fetch_definition.side_effect = [
        wiz.exception.RequestNotFound("Error"),
        wiz.definition.Definition(
            {"identifier": "bar"},
            path="/path/to/registry/bar/bar.json"
        ),
        wiz.definition.Definition(
            {
                "identifier": "baz",
                "version": "2.5.1",
                "description": "test",
            },
            path="/path/to/registry/baz/baz-2.5.1.json"
        ),
    ]

    wiz.registry.install_to_path(
        mocked_definitions, temporary_directory, overwrite=True
    )

    mocked_fetch_definition_mapping.assert_called_once_with(
        [temporary_directory]
    )

    assert mocked_fetch_definition.call_count == 3
    mocked_fetch_definition.assert_any_call("foo==0.1.0", "__MAPPING__")
    mocked_fetch_definition.assert_any_call("bar", "__MAPPING__")
    mocked_fetch_definition.assert_any_call("baz==2.5.1", "__MAPPING__")

    mocked_remove.assert_called_once_with(
        "/path/to/registry/baz/baz-2.5.1.json"
    )

    assert mocked_export_definition.call_count == 2
    mocked_export_definition.assert_any_call(
        temporary_directory, mocked_definitions[0], overwrite=True
    )
    mocked_export_definition.assert_any_call(
        "/path/to/registry/baz", mocked_definitions[2], overwrite=True
    )

    logger.info.assert_called_once_with(
        "Successfully installed 2 definition(s) to registry '{}'."
        .format(temporary_directory)
    )


def test_install_to_path_no_content(
    temporary_directory, mocked_definitions, mocked_fetch_definition_mapping,
    mocked_fetch_definition, mocked_export_definition, mocked_remove, logger,
):
    """Fail to install definitions when no new content available."""
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_fetch_definition.side_effect = mocked_definitions

    with pytest.raises(wiz.exception.InstallNoChanges) as error:
        wiz.registry.install_to_path(mocked_definitions, temporary_directory)

    mocked_fetch_definition_mapping.assert_called_once_with(
        [temporary_directory]
    )

    assert mocked_fetch_definition.call_count == 3
    mocked_fetch_definition.assert_any_call("foo==0.1.0", "__MAPPING__")
    mocked_fetch_definition.assert_any_call("bar", "__MAPPING__")
    mocked_fetch_definition.assert_any_call("baz==2.5.1", "__MAPPING__")

    mocked_remove.assert_not_called()
    mocked_export_definition.assert_not_called()
    logger.info.assert_not_called()

    assert "InstallNoChanges: Nothing to install." in str(error)



# :coding: utf-8

"""Wiz documentation build configuration file."""

import os
import re
import sys

# -- General ------------------------------------------------------------------
# Inject source onto path so that autodoc can find it by default, but in such a
# way as to allow overriding location.
sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "source"))
)

# Extensions.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_click.ext",
    "lowdown"
]

# Add local extensions.
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '_extension'))
extensions.append("code_block")

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = u"Wiz"
copyright = u"2018, The Mill"

# Version
with open(
    os.path.join(
        os.path.dirname(__file__), "..", "source",
        "wiz", "_version.py"
    )
) as _version_file:
    _version = re.match(
        r".*__version__ = \"(.*?)\"", _version_file.read(), re.DOTALL
    ).group(1)

version = _version
release = _version

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_template"]

# A list of prefixes to ignore for module listings.
modindex_common_prefix = [
    "wiz."
]

# -- HTML output --------------------------------------------------------------

html_theme = "sphinx_rtd_theme"

html_static_path = ["_static"]

# If True, copy source rst files to output for reference.
html_copy_source = True


# -- Autodoc ------------------------------------------------------------------

autodoc_default_flags = ["members", "undoc-members", "inherited-members"]
autodoc_member_order = "bysource"


def autodoc_skip(app, what, name, obj, skip, options):
    """Don"t skip __init__ method for autodoc."""
    if name == "__init__":
        return False

    return skip


# -- Intersphinx --------------------------------------------------------------

intersphinx_mapping = {
    "python": ("http://docs.python.org/", None),
    "packaging": ("https://packaging.pypa.io/en/stable/", None),
    "click": ("https://click.palletsprojects.com/en/7.x/", None),
    "sawmill": ("https://sawmill.readthedocs.io/en/stable/", None)
}


# -- Setup --------------------------------------------------------------------

def setup(app):
    """Setup *app*."""
    app.connect("autodoc-skip-member", autodoc_skip)

    # Ensure custom stylesheet added, even when on Read The Docs server where
    # html_style setting is ignored.
    app.add_stylesheet("style.css")



# :coding: utf-8

import getpass
import os
import socket

import pytest

import wiz.config
import wiz.environ


@pytest.fixture()
def initial_environment(mocker):
    """Mock environment to use when fetching config."""
    mocker.patch.object(socket, "gethostname", return_value="__HOSTNAME__")
    mocker.patch.object(getpass, "getuser", return_value="__USER__")
    mocker.patch.object(os.path, "expanduser", return_value="__HOME__")

    # Reset configuration.
    wiz.config.fetch(refresh=True)


@pytest.mark.usefixtures("initial_environment")
def test_initiate():
    """Return initial data mapping."""
    assert wiz.environ.initiate() == {
        "USER": "__USER__",
        "HOME": "__HOME__",
        "HOSTNAME": "__HOSTNAME__",
        "PATH": os.pathsep.join([
            "/usr/local/sbin",
            "/usr/local/bin",
            "/usr/sbin",
            "/usr/bin",
            "/sbin",
            "/bin",
        ])
    }


@pytest.mark.usefixtures("initial_environment")
def test_initiate_with_config_passthrough(monkeypatch):
    """Return initial data mapping with passthrough environment variables."""
    monkeypatch.setenv("ENVIRON_TEST1", "VALUE")
    monkeypatch.delenv("ENVIRON_TEST2", raising=False)

    # Add passthrough list in config.
    config = wiz.config.fetch()
    config["environ"]["passthrough"] = ["ENVIRON_TEST1", "ENVIRON_TEST2"]

    assert wiz.environ.initiate() == {
        "ENVIRON_TEST1": "VALUE",
        "USER": "__USER__",
        "HOME": "__HOME__",
        "HOSTNAME": "__HOSTNAME__",
        "PATH": os.pathsep.join([
            "/usr/local/sbin",
            "/usr/local/bin",
            "/usr/sbin",
            "/usr/bin",
            "/sbin",
            "/bin",
        ])
    }


@pytest.mark.usefixtures("initial_environment")
def test_initiate_with_config_initial():
    """Return initial data mapping with passthrough environment variables."""
    # Add initial mapping in config.
    config = wiz.config.fetch()
    config["environ"]["initial"] = {"ENVIRON_TEST1": "VALUE"}

    assert wiz.environ.initiate() == {"ENVIRON_TEST1": "VALUE"}


@pytest.mark.usefixtures("initial_environment")
def test_initiate_with_mapping():
    """Return mapping with initial data mapping."""
    assert wiz.environ.initiate(
        mapping={
            "HOSTNAME": "__OTHER_HOSTNAME__",
            "KEY": "VALUE"
        }
    ) == {
        "USER": "__USER__",
        "HOME": "__HOME__",
        "HOSTNAME": "__OTHER_HOSTNAME__",
        "PATH": os.pathsep.join([
            "/usr/local/sbin",
            "/usr/local/bin",
            "/usr/sbin",
            "/usr/bin",
            "/sbin",
            "/bin",
        ]),
        "KEY": "VALUE"
    }


def test_sanitize():
    """Return sanitized environment mapping"""
    assert wiz.environ.sanitize({
        "PLUGINS_A": "/path/to/plugin1:/path/to/plugin2:${PLUGINS_A}",
        "PLUGINS_B": "${PLUGINS_B}:/path/to/plugin1:/path/to/plugin2",
        "PLUGINS_C": "/path/to/plugin1:${PLUGINS_C}:/path/to/plugin2",
        "PLUGINS_D": "${PLUGINS_C}:${PLUGINS_D}",
        "KEY1": "HELLO",
        "KEY2": "${KEY1} WORLD!",
        "KEY3": "${UNKNOWN}"
    }) == {
        "PLUGINS_A": "/path/to/plugin1:/path/to/plugin2",
        "PLUGINS_B": "/path/to/plugin1:/path/to/plugin2",
        "PLUGINS_C": "/path/to/plugin1:/path/to/plugin2",
        "PLUGINS_D": "/path/to/plugin1:/path/to/plugin2",
        "KEY1": "HELLO",
        "KEY2": "HELLO WORLD!",
        "KEY3": "${UNKNOWN}"
    }

    assert wiz.environ.sanitize({
        "PLUGINS_A": "/path/to/plugin1:/path/to/plugin2:$PLUGINS_A",
        "PLUGINS_B": "$PLUGINS_B:/path/to/plugin1:/path/to/plugin2",
        "PLUGINS_C": "/path/to/plugin1:$PLUGINS_C:/path/to/plugin2",
        "PLUGINS_D": "$PLUGINS_C:$PLUGINS_D",
        "KEY1": "HELLO",
        "KEY2": "$KEY1 WORLD!",
        "KEY3": "$UNKNOWN"
    }) == {
        "PLUGINS_A": "/path/to/plugin1:/path/to/plugin2",
        "PLUGINS_B": "/path/to/plugin1:/path/to/plugin2",
        "PLUGINS_C": "/path/to/plugin1:/path/to/plugin2",
        "PLUGINS_D": "/path/to/plugin1:/path/to/plugin2",
        "KEY1": "HELLO",
        "KEY2": "HELLO WORLD!",
        "KEY3": "$UNKNOWN"
    }


def test_contains():
    """Indicate whether *text* contains a reference to variable."""
    assert wiz.environ.contains("$HOME/to/data", "HOME") is True
    assert wiz.environ.contains("${HOME}/to/data", "HOME") is True
    assert wiz.environ.contains("$PATH/to/data", "HOME") is False


@pytest.mark.parametrize("text, expected", [
    ("$HOME/to/data", "/usr/people/me/to/data"),
    ("${HOME}/to/data", "/usr/people/me/to/data"),
    ("$HOME/to/$OS/data", "/usr/people/me/to/centos/data"),
    ("$HOME/to/${OS}/data", "/usr/people/me/to/centos/data"),
], ids=[
    "one-variable",
    "one-variable-with-curly-brackets",
    "several-variables",
    "several-variables-mixed",
])
def test_substitute(text, expected):
    """Substitute all environment variables in *text* from environment."""
    environment = {
        "HOME": "/usr/people/me",
        "OS": "centos"
    }
    assert wiz.environ.substitute(text, environment) == expected
